# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.

# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
import random
from PIL import ImageFilter

import torch
import torch.nn as nn


class SimSiam(nn.Module):
    """
    Build a SimSiam model.
    """

    def __init__(self, base_encoder, dim=1024, pred_dim=512):
        super(SimSiam, self).__init__()
        self.dim = dim  # dim: feature dimension (default: 128)
        self.pred_dim = pred_dim  # pred_dim: hidden dimension of the predictor (default: 512)
        self.online_network = base_encoder(num_classes=self.dim, zero_init_residual=True)
        self.target_network = base_encoder(num_classes=self.dim, zero_init_residual=True)

        prev_dim = self.online_network.fc.weight.shape[1]
        self.online_network.fc = nn.Sequential(
            nn.Linear(prev_dim, prev_dim, bias=False),
            nn.BatchNorm1d(prev_dim),
            nn.ReLU(inplace=True),  # first layer
            nn.Linear(prev_dim, prev_dim, bias=False),
            nn.BatchNorm1d(prev_dim),
            nn.ReLU(inplace=True),  # second layer
            nn.Linear(prev_dim, self.dim, bias=False),
            nn.BatchNorm1d(self.dim, affine=False),  # third layer

            nn.Linear(self.dim, self.pred_dim, bias=False),
            nn.BatchNorm1d(self.pred_dim),
            nn.ReLU(inplace=True),  # fourth layer
            nn.Linear(self.pred_dim, self.dim))  # output layer

        self.target_network.fc = nn.Sequential(
            nn.Linear(prev_dim, prev_dim, bias=False),
            nn.BatchNorm1d(prev_dim),
            nn.ReLU(inplace=True),  # first layer
            nn.Linear(prev_dim, prev_dim, bias=False),
            nn.BatchNorm1d(prev_dim),
            nn.ReLU(inplace=True),  # second layer
            nn.Linear(prev_dim, self.dim, bias=False),
            nn.BatchNorm1d(self.dim, affine=False),  # third layer

            nn.Linear(self.dim, self.pred_dim, bias=False),
            nn.BatchNorm1d(self.pred_dim),
            nn.ReLU(inplace=True),  # fourth layer
            nn.Linear(self.pred_dim, self.dim))  # output layer

    def forward(self, img):
        predictions_from_img = self.online_network(img)
        targets_to_img = self.target_network(img)

        return predictions_from_img, targets_to_img


class TwoCropsTransform:
    """Take two random crops of one image as the query and key."""

    def __init__(self, base_transform):
        self.base_transform = base_transform

    def __call__(self, x):
        q = self.base_transform(x)
        k = self.base_transform(x)
        return [q, k]


class GaussianBlur(object):
    """Gaussian blur augmentation in SimCLR https://arxiv.org/abs/2002.05709"""

    def __init__(self, sigma=[.1, 2.]):
        self.sigma = sigma

    def __call__(self, x):
        sigma = random.uniform(self.sigma[0], self.sigma[1])
        x = x.filter(ImageFilter.GaussianBlur(radius=sigma))
        return x



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
import os
import logging

import torch
import torchvision
import torch.nn as nn
import torch.nn.functional as F
import torchvision.models as models
from torch.utils.data.dataloader import DataLoader


class BYOL(nn.Module):
    def __init__(self, base_encoder, dim=128, mlp=True):
        super(BYOL, self).__init__()
        self.dim = dim  # dim: feature dimension (default: 128)
        self.online_network = base_encoder(num_classes=self.dim)
        self.target_network = base_encoder(num_classes=self.dim)

        if mlp:  # hack: brute-force replacement
            dim_mlp = self.online_network.fc.weight.shape[1]
            self.online_network.fc = nn.Sequential(
                nn.Linear(dim_mlp, dim_mlp),
                nn.BatchNorm1d(dim_mlp),
                nn.ReLU(),
                nn.Linear(dim_mlp, self.dim),

                nn.Linear(self.dim, self.dim),
                nn.BatchNorm1d(self.dim),
                nn.ReLU(),
                nn.Linear(self.dim, self.dim))

            self.target_network.fc = nn.Sequential(
                nn.Linear(dim_mlp, dim_mlp),
                nn.BatchNorm1d(dim_mlp),
                nn.ReLU(),
                nn.Linear(dim_mlp, self.dim),

                nn.Linear(self.dim, self.dim),
                nn.BatchNorm1d(self.dim),
                nn.ReLU(),
                nn.Linear(self.dim, self.dim))

    def forward(self, img):
        predictions_from_img = self.online_network(img)
        targets_to_img = self.target_network(img)

        return predictions_from_img, targets_to_img


class BYOLTrainer:
    def __init__(self, model, optimizer, args, **params):
        self.model = model
        self.optimizer = optimizer
        self.args = args

    @torch.no_grad()
    def _update_target_network_parameters(self):
        """
        Momentum update of the key encoder
        """
        for param_q, param_k in zip(self.model.online_network.parameters(), self.model.target_network.parameters()):
            param_k.data = param_k.data * self.args.m + param_q.data * (1. - self.args.m)

    @staticmethod
    def regression_loss(x, y):
        x = F.normalize(x, dim=1)
        y = F.normalize(y, dim=1)
        return 2 - 2 * (x * y).sum(dim=-1)

    def initializes_target_network(self):
        # init momentum network as encoder net
        for param_q, param_k in zip(self.model.online_network.parameters(), self.model.target_network.parameters()):
            param_k.data.copy_(param_q.data)  # initialize
            param_k.requires_grad = False  # not update by gradient

    def train(self, train_dataset):
        # drop_last舍去不足一个bach_size的样本
        train_loader = DataLoader(train_dataset, batch_size=self.args.batch_size, num_workers=self.args.num_workers, drop_last=False, shuffle=True)
        niter = 0
        self.initializes_target_network()

        for epoch_counter in range(self.args.epochs):
            loss = 0.
            for batch_view_1, batch_view_2 in train_loader:
                batch_view_1 = batch_view_1.to(self.args.device)
                batch_view_2 = batch_view_2.to(self.args.device)

                loss = self.L2_loss(batch_view_1, batch_view_2)
                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()

                self._update_target_network_parameters()  # update the key encoder
                niter += 1

            logging.info("End of epoch {}  loss:{}".format(epoch_counter, loss))

            # save checkpoints
            if (epoch_counter + 1) % self.args.save_freq == 0:
                save_checkpoint({'epoch': epoch_counter + 1, 'arch': self.args.arch, 'optimizer': self.optimizer.state_dict(), 'state_dict': self.model.state_dict()},
                                self.args.log_dir,
                                filename='epoch_{:04d}.pth.tar'.format(epoch_counter + 1))
                logging.info('epoch{:04d}.pth.tar saved!'.format(epoch_counter + 1))

    def L2_loss(self, batch_view_1, batch_view_2):

        # compute query feature
        predictions_from_view_1 = self.model.online_network(batch_view_1)
        predictions_from_view_2 = self.model.online_network(batch_view_2)

        # compute key features
        with torch.no_grad():
            self.model.target_network.fc[3].register_forward_hook(get_activation('projections_from_view_1'))
            self.model.target_network(batch_view_1)
            targets_to_view_1 = torch.squeeze(activation['projections_from_view_1'])

            self.model.target_network.fc[3].register_forward_hook(get_activation('projections_from_view_2'))
            self.model.target_network(batch_view_2)
            targets_to_view_2 = torch.squeeze(activation['projections_from_view_2'])

        loss = self.regression_loss(predictions_from_view_1, targets_to_view_1)
        loss += self.regression_loss(predictions_from_view_2, targets_to_view_2)
        return loss.mean()


def save_checkpoint(state, log_dir, filename):
    filename_path = os.path.join(log_dir, filename)
    torch.save(state, filename_path)
    print(filename + ' has been saved.')


activation = {}


def get_activation(name):
    def hook(model, input, output):
        activation[name] = output

    return hook


if __name__ == '__main__':
    model_names = sorted(name for name in models.__dict__
                         if name.islower() and not name.startswith("__")
                         and callable(models.__dict__[name]))

    model = BYOL(models.__dict__['resnet18'], 128, True).to(torch.device('cpu'))
    print(model)
    print("__" * 50)
    print(model.online_network)
    print("__" * 50)
    print(model.target_network)



from setuptools import find_packages, setup

with open('README.md', 'r', encoding='utf-8') as fp:
    long_description = fp.read()

setup(
    name='damilu_baseline',
    version='0.0.1',
    author='Hopetree',
    author_email='zlwork2014@163.com',
    description='A web tip of Django',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com',
    keywords='damilu',
    packages=find_packages(),
    include_package_data=True,
    python_requires='>=3.5',
    classifiers=[
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License'
    ],
)